"use strict";

const _            = require('lodash');
const EventEmitter = require('events');
const Promise      = require('bluebird');
const debug        = require('debug')('eventbus:bus');
const amqplib      = require('amqplib');
const isFatalError = require('amqplib/lib/connection').isFatalError;
const domain       = require('domain');

const RECONNET_INTERVAL_DEFAULT = 1; // in seconds

class Bus extends EventEmitter {

    constructor(microServiceName, options) {
        super();
        this.url               = options.url;
        this.microServiceName  = microServiceName;
        this.reconnectInterval = (options.reconnectInterval || RECONNET_INTERVAL_DEFAULT) * 1000;
        this.subscribers = {};
        this.initializing = false;
        this._reSubscribe = this._reSubscribe.bind(this);
        this._reconnect = this._reconnect.bind(this);
    }

    _createErrorHandler() {
        if (this.errorHandlerCreated) {
            return;
        }

        this.errorHandlerCreated = true;
        this.on('error', (err) => {
            if (isFatalError(err)) { // reconnect only on fatal error
                this._reconnect();
            }
        });
    }

    async quit() {
        await this.connection.close();
        this.subscribers = {};
        debug('bus closed');
    }

    init() {
        this.initializing = true;
        this._createErrorHandler();

        this.openPromise = amqplib.connect(this.url);
        this.openPromise
            .then((newConnection) => {
                debug('new connection created');
                if (this.reconnect && this.connection) {
                    this.publishChannel = undefined;
                    debug('reconnected');
                }
                this.reconnect  = false;
                this.connection = newConnection;

                debug(`ready`);
                this.emit('ready');

                this.connection.on('error', (err) => {
                    this.emit('error', err);
                });

                this.connection.on('close', () => {
                    debug(`connection close event`);
                    this.emit('close');
                    this._reconnect();
                });

                this.connection.on('blocked', () => {
                    debug('connection blocked');
                });

                this.connection.on('unblocked', () => {
                    debug('connection unblocked');
                });

                return this._reSubscribe();
            })
            .finally(() => {
                this.initializing = false;
            })
            .catch((err) => {
                this.emit('error', err);
            })
            .done();
    }

    _reconnect() {
        if (this.initializing) {
            debug('not reconnecting because initialization is still in process');
            return;
        }

        debug(`reconnecting in: ${this.reconnectInterval / 1000} seconds because of fatal error`);
        this.reconnect = true;
        setTimeout(() => {
            debug(`reconnecting`);
            this.init();
        }, this.reconnectInterval);
    }

    _reSubscribe() {
        return Promise.map(_.values(this.subscribers), (eventSubscribers) => {
            // `Promise.each` for guarantee of order for subscribers
            return Promise.each(eventSubscribers, ({eventName, handler, options}) => {
                return this.subscribe(eventName, handler, options, true);
            });
        });
    }

    publish(eventName, msg) {
        return Promise.resolve()
            .then(() => {
                if (!this.publishChannel) {
                    return this._createPublishChannel();
                }
            })
            .then(() => {
                return this.publishChannel.assertExchange(eventName, 'fanout', { durable: true });
            })
            .then(() => {
                return this.publishChannel.publish(eventName, 'my-routing-key', new Buffer(this._serialize(msg)));
            });
    }

    subscribe(eventName, handler, options = {}, restarting) {
        if (!restarting) {
            this.subscribers[eventName] = this.subscribers[eventName] || [];
            this.subscribers[eventName].push({eventName, handler, options});
        }
        const subscriberEmitter = new EventEmitter();
        subscriberEmitter.on('error', () => {});
        const waitForAck = options.waitForAck || false;
        const concurrencyLimit =  options.concurrencyLimit || false;
        const messageTtl = options.ttl || false;
        const timeout = options.timeout;
        const ackOnTimeout = options.ackOnTimeout;

        const serviceRole = options.serviceRole ? `${options.serviceRole}_` : '';
        const additionalIdentifier = options.additionalIdentifier ? `${options.additionalIdentifier}_` : '';
        const queueName = `${this.microServiceName}_${serviceRole}${additionalIdentifier}${eventName}`;
        Promise.resolve()
            .then(() => {
                return this._getSubscribeChannel();
            })
            .then((ch) => {
                return ch.assertExchange(eventName, 'fanout', { durable: true })
                    .then(() => {
                        const queueOptions = { durable: true };
                        if (messageTtl) {
                            queueOptions.messageTtl = messageTtl;
                        }
                        return ch.assertQueue(queueName, queueOptions)
                            .then(() => {
                                debug(`queue: ${queueName} asserted`);
                                return ch.bindQueue(queueName, eventName);
                            })
                            .then(() => {
                                return ch.prefetch(concurrencyLimit);
                            })
                            .then(() => {
                                debug(`queue: ${queueName} concurrencyLimit: ${concurrencyLimit} was set`);
                                return ch.consume(queueName, (msg) => {
                                    const d = domain.create();
                                    d.on('error', (err) => {
                                        subscriberEmitter.emit('error', err);
                                    });
                                    d.run(() => {
                                        let finished = false;
                                        if (waitForAck) {
                                            if (timeout) {
                                                setTimeout(() => {
                                                    if (!finished) {
                                                        finished = true;
                                                        subscriberEmitter.emit('error', new Error('Handler timedout'));
                                                        if (ackOnTimeout) {
                                                            ch.ack(msg);
                                                        } else {
                                                            ch.nack(msg);
                                                        }
                                                    }
                                                }, timeout);
                                            }
                                        } else {
                                            ch.ack(msg);
                                        }
                                        handler(this._deserialize(msg.content.toString()))
                                            .then(() => {
                                                if (!finished && waitForAck) {
                                                    finished = true;
                                                    ch.ack(msg);
                                                } else {
                                                    finished = true;
                                                }
                                            }, (err) => {
                                                subscriberEmitter.emit('error', err);
                                                if (!finished && waitForAck) {
                                                    finished = true;
                                                    ch.nack(msg);
                                                } else {
                                                    finished = true;
                                                }
                                            });
                                    });
                                });
                            });
                    });
            })
            .catch((err) => {
                subscriberEmitter.emit('error', { connection: err });
            })
            .done();

        return subscriberEmitter;
    }

    _createPublishChannel() {
        return this.openPromise
            .then((conn) => {
                return conn.createChannel();
            })
            .then((ch) => {
                this.publishChannel = ch;
            });
    }

    _getSubscribeChannel() {
        return this.openPromise
            .then((conn) => {
                return conn.createChannel();
            });
    }

    _serialize(body) {
        return JSON.stringify(body);
    }

    _deserialize(content) {
        return JSON.parse(content);
    }


}


module.exports = Bus;
